//  ------------------------------------------------------------------------
//@ BtnAni.pas
//  TBtnAni VCL Component
//  Copyright (c) Cole Joe Taiwan 1998 All rights reserved.
//  email: kchou@adc.iii.org.tw
//  -   Description
//      TBtnAni is a Animation Button object inherits from TImageAni with
//      the following features:
//      1.  Animation High Light effect:
//          There are two kinds of type supported, bbBtn and bbChkBox
//          bbBtn:      FrameCount should be >=4, Frame ID more than 4
//                      will be treated as High Light feature for looping play
//          bbChkBox:   FrameCount should be >=6, Frame ID more than 6
//                      will be treated as High Light feature for looping play
//    	2.  OnClick sound and High Light sound supported by setting the properties
//			of SndOnClicking and SndHightLight resp.
//      3.  Like TSpeedButton, this component support With GroupIndex feature
//      4.  manipulate misc mouse event
//          OnMouseMove, OnMouseDown, OnMouseUp, OnMouseEnter, OnMouseLeave
//          Automatically update Button appearance of High light, down/up,
//          enable/disable, check/uncheck state
//  -   Frame ID definition:
//      -   For bbBtn
//          0:  Disable
//          1:  Button Down
//          2:  Button High Light
//          3:  Button Up
//          Frame ID more than 4 will be treated as High Light feature for
//          looping play
//      -   For bbChkBox
//          0:  Button Check Off
//          1:  Button Check Off High Light
//          2:  Button Check Off and Disable
//          3:  Button Check On
//          4:  Button Check On High Light
//          5:  Button Check On and Disable
//          Frame ID more than 6 will be treated as High Light feature for
//          looping play
//  -   When OnClick event Happened:
//      -   For bbChkBox
//          OnClick happened on MouseDown
//      -   For bbBtn
//          OnClick happened on MouseUp
//  -   980910
//      Add Tag1, Tag2 properties
//  -   980909
//      1.  Add GroupIndex property and UpdateExclusive
//      2.  Add Down property
//          Btn can Down only if, Btn is bbBtn and GroupIndex<>0
//          if Down=true and _SameGroupIndexSiblingCount>0 then
//          Mouse Message won't take effect
//  -   980908
//      1.  Override property FrameCount by add an override method
//          procedure _SetTotalFrame(Value : Integer); override;
//      2.  Add new Callback events
//          OnMouseEnter, OnMouseLeave
//          -   OnMouseEnter will be invoke under the following conditions
//              a.  Mouse Move on Control and
//              b.  Control is Enabled and
//              c.  Not any TBtnAni object is dragging
//          -   OnMouseLeave will be invoke under the following conditions
//              a.  Mouse Leave Control and
//              b.  Control is Enabled and
//              c.  Not any TBtnAni object is dragging
//  -   980907
//      Modified TBtnAni.Init and Add m_sHintOn, m_sHintOff, m_sFileWav
//  ------------------------------------------------------------------------
unit BtnAni;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages,
  Classes, Graphics, Controls, Forms,
  ImageAni;
const
  ///////////////////////////////////////////////////////////////////
  // Glyph index
  ///////////////////////////////////////////////////////////////////
  // TBtnState = (bbBtnDisable, bbBtnDn, bbBtnHLight, bbBtnUp);
  // for m_BtnType=bbBtn= -- Button
  bbBtnDisable=0;
  bbBtnDn=1;
  bbBtnHLight=2;
  bbBtnUp=3;
  // TChkBoxState = (bbChkBoxOff, bbChkBoxOffHLight, bbChkBoxOffDisable,
  //                 bbChkBoxOn,  bbChkBoxOnHLight,  bbChkBoxOnDisable);
  bbChkBoxOff=0;
  bbChkBoxOffHLight=1;
  bbChkBoxOffDisable=2;
  bbChkBoxOn=3;
  bbChkBoxOnHLight=4;
  bbChkBoxOnDisable=5;
type
TBtnAniType = (bbBtn, bbChkBox);
TBtnAni = class(TImageAni) // [
// New member
private
    FDown: Boolean;
    FGroupIndex: Integer;
    m_bDragging: boolean; // MyBmpAni been Clicked Down
    m_bEnabled: boolean;    // is Control Enabled
    m_bChecked: boolean;    // is Control Checked, only apply on bbChkBox
    m_bHLight: boolean;     // is Control High Light
    m_BtnType: TBtnAniType;
    m_sFileSndHiLight: string; // High Light WAV File
    m_sFileSndClicking: string;// Clicking WAV File
    m_sHintOn: string; // for Check Btn is On Hint, for Button is Hint
    m_sHintOff: string; // for Check Btn is Off Hint
    m_sTag1, m_sTag2: string;
    _OnMouseEnter: TNotifyEvent;
    _OnMouseLeave: TNotifyEvent;
    procedure _SetTotalFrame(Value : Integer); override;
    procedure _PlaySnd(sFile: string);
    procedure _SetHintOn(sHintOn: string);
    procedure _SetHintOff(sHintOff: string);
    //  CMMouseEnter will raise OnMouseEnter callback event
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    //  CMMouseLeave will raise OnMouseLeave callback event
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure UpdateExclusive; // Update Same GroupIndex Sibling, Same as TSpeedButton.UpdateExclusive;
    function _SameGroupIndexSiblingCount: integer;
    procedure SetGroupIndex(Value: integer);
    procedure SetDown(Value: boolean);
protected
    //procedure Paint; override;
    //procedure TimeHit(Sender : TObject);
    //  override parent's MouseDown/MouseMove/MouseUp operations
    procedure MouseDown(Button: TMouseButton;
                        Shift: TShiftState;
                        X, Y: Integer); override;
    procedure MouseUp  (Button: TMouseButton;
                        Shift: TShiftState;
                        X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState;
                        X, Y: Integer); override;
    // Button Left Down/Move/BtnUp operation
    //function OnBtnDn(Button: TMouseButton; Shift: TShiftState; X, Y: Integer): boolean;
    //function OnBtnUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer): boolean;
    //procedure OnBtnMove(Shift: TShiftState; X, Y: Integer);
public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Click; override;
  	//////////////////////////
	function    Init(nBtnType: TBtnAniType;
				     sFileBmp: string;
                     sFileSndHiLight, sFileSndClick: string;
				     nFrameCount: integer;
				     nX, nY: integer;		// position
                     sHintOn, sHintOff: string): boolean;
    function    IsChkBox: boolean;
    function    IsBtn: boolean;
    function    IsEnabled : boolean;
    procedure   DrawBtnUp; // Draw BtnUp Button appearance
    procedure   SetEnabled(Value: boolean);
    function    IsChecked : boolean;
    procedure   SetChecked(Value: boolean);
    function    IsHLight: boolean;
    procedure   SetHLight;
    procedure   SetType(nBtnType: TBtnAniType);
    procedure   SetSndHighLight(sFileName: string);
    procedure   SetSndClicking(sFileName: string);
published
    property Interval;
    property FrameCount;
    //property StartFrame default 0;
    //property StopFrame default 0;
    property Frame default 0;
    property FileName;
    property Play;
    property Reverse;
    property Loop;
    property TransparentColor;
    property Height default 30;
    property Width default 30;
    property ShowHint default false;
    property OnChangeFrame;
    //property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
  	property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseEnter:TNotifyEvent  read _OnMouseEnter
                                        write _OnMouseEnter;
    property OnMouseLeave:TNotifyEvent  read _OnMouseLeave
                                        write _OnMouseLeave;
    property Visible;
	property Checked: boolean           read IsChecked
                                        write SetChecked
                                        default true;
	property Enabled: boolean           read IsEnabled
                                        write SetEnabled
                                        default true;
    property BtnType: TBtnAniType       read m_BtnType
                                        write SetType
                                        default bbBtn;
    property SndOnClicking: string      read m_sFileSndClicking
                                        write SetSndClicking;
    property SndHightLight: string      read m_sFileSndHiLight
                                        write SetSndHighLight;
    //  HintOn: for bbBtn,      HintOn is Hint
    //          for bbChkBox,   HintOn is Hint on Toggle ON
    property HintOn: string             read m_sHintOn
                                        write _SetHintOn;
    //  HintOff:for bbBtn,      HintOff is neglected
    //          for bbChkBox,   HintOff is Hint on Toggle OFF
    property HintOff: string            read m_sHintOff
                                        write _SetHintOff;
    //  980909
	property Down: boolean              read FDown
                                        write SetDown
                                        default FALSE;
    //  GroupIndex: not allow in run time
    property GroupIndex: Integer        read FGroupIndex
                                        write SetGroupIndex
                                        default 0;
    property Tag1: string               read m_sTag1 write m_sTag1;
    property Tag2: string               read m_sTag2 write m_sTag2;
  end; // ] TBtnAni
implementation
uses
    MyMacro,
    mmsystem; // for sndPlaySound
var
    s_bOnDragging: boolean; // on Dragging on BtnAni
const
    MAX_NUM_BTN_STATE=4;
    MAX_NUM_CHK_STATE=6;
{$R BtnAni.res}
function IsLoading(Sender: TComponent): boolean;
begin
    Result:=csLoading in Sender.ComponentState;
end;
function IsSpace(c: char): boolean;
begin
    Result:=FALSE;
    if ((Ord(c)>=$09) and (Ord(c)<=$0D)) or (Ord(c)=$20) then
    	Result:=TRUE;
end;
//  -------------------------------------------------------------------
// @ IsCharNonsense
//   function IsCharNonsense(c: char) : boolean;
//	 #define IsSpace(c) (((int)c>=0x09 && (int)c<=0x0D) || (int)c==0x20)
//  -------------------------------------------------------------------
function IsCharNonsense(c: char) : boolean;
begin
    Result:=FALSE;
    if  IsSpace(c) or (Ord(c)<32) then
    	Result:=TRUE;
end; // IsCharNonsense
//  -------------------------------------------------------------------
//@ IsStrNonsense
//  function IsStrNonsense(s: string) : boolean;
//  Check if s is nonsense
//  -------------------------------------------------------------------
function IsStrNonsense(s: string) : boolean;
var
    i, nLen : integer;
begin // [
    if  s='' then begin
        Result:=TRUE; exit
    end;
    nLen:=Length(s);
    for i:=1 to nLen do begin // [
        if  not IsCharNonsense(s[i]) then begin
            Result:=FALSE; exit
        end;
    end; //] for i:=1 to nLen
    Result:=TRUE;
end; // ] IsStrNonsense
constructor TBtnAni.Create(AOwner: TComponent);
begin // [
    inherited Create(AOwner);
    m_bDragging:=false;
    m_BtnType:=bbBtn;
    m_bEnabled:=true;
    m_bChecked:=false;
    m_bHLight:=false;
    m_sFileSndHiLight:='';
    m_sFileSndClicking:='';
    m_sHintOn:=''; m_sHintOff:='';
    Interval:=150;
end; // ] TBtnAni.Create
destructor TBtnAni.Destroy;
begin
    inherited Destroy;
end;
procedure _SetAndChkSndFile(var sFileWavDest: string; sFileWav: string);
var
    sExt: string;
begin // [
    sFileWavDest:=sFileWav;
    //  Validate *.WAV file
    if  IsStrNonsense(sFileWavDest) then begin
        sFileWavDest:=''; exit;
    end;
    if  not FileExists(sFileWavDest) then begin
        sFileWavDest:=''; exit;
    end;
    if  Length(sFileWavDest)<=4 then begin
        sFileWavDest:=''; exit; // at least "1.WAV", 5 chars
    end;
    sExt:=ExtractFileName(sFileWavDest);
    sExt:=Copy(sExt, Length(sExt)-2, 3); // Check last 3 char is WAV?
    if  UpperCase(sExt)<>'WAV' then
        sFileWavDest:='';
end; // ] _SetAndChkSndFile
//  Set Clicking Sound FileName
procedure TBtnAni.SetSndClicking(sFileName: string);
begin // [
    _SetAndChkSndFile(m_sFileSndClicking, sFileName);
end; // ] TBtnAni.SetSndClicking
//  Set High Light Sound FileName
procedure TBtnAni.SetSndHighLight(sFileName: string);
begin // [
    _SetAndChkSndFile(m_sFileSndHiLight, sFileName);
end; // ] TBtnAni.SetSndHighLight
procedure TBtnAni.SetType(nBtnType: TBtnAniType);
begin // [
    if  m_BtnType=nBtnType then exit;
    m_BtnType:=nBtnType;
    _SetTotalFrame(FrameCount);
end; // ] TBtnAni.SetType
procedure TBtnAni._SetTotalFrame(Value : Integer);
begin // [
    // Set StartFrame, StopFrame, Frame, FrameCount
    if  m_BtnType=bbBtn then begin
        if  Value<MAX_NUM_BTN_STATE then
            Value:=MAX_NUM_BTN_STATE;
    end;
    if  m_BtnType=bbChkBox then begin
        if  Value<MAX_NUM_CHK_STATE then
            Value:=MAX_NUM_CHK_STATE;
    end;
    inherited _SetTotalFrame(Value);
    if  m_BtnType=bbBtn then begin
        if  Value>MAX_NUM_BTN_STATE then begin
            StartFrame:=MAX_NUM_BTN_STATE; StopFrame:=Value-1;
        end;
    end;
    if  m_BtnType=bbChkBox then begin
        if  Value>MAX_NUM_CHK_STATE then begin
            StartFrame:=MAX_NUM_CHK_STATE; StopFrame:=Value-1;
        end;
    end;
    if  m_BtnType=bbBtn then    Frame:=bbBtnUp;
    if  m_BtnType=bbChkBox then Frame:=bbChkBoxOn;
end; // ] TBtnAni._SetTotalFrame
//  -----------------------------------------------------------------------------
//@ TBtnAni.Init
//  Initialize TBtnAni data member
//  -----------------------------------------------------------------------------
function TBtnAni.Init(nBtnType: TBtnAniType;
				        sFileBmp: string;
                        sFileSndHiLight, sFileSndClick: string;
				        nFrameCount: integer;
				        nX, nY: integer;		// position
                        sHintOn, sHintOff: string): boolean;
begin // [
    Result:=true;
    if( ((nBtnType<>bbBtn) and (nBtnType<>bbChkBox)) or
        (not FileExists(sFileBmp))) then begin // [
        Result:=false; exit;
    end; // ] illegal initial value
    //  Set m_BtnType
    m_BtnType:=nBtnType;
    //  Set Loop, FileName, FrameCount, Left, Top
    Loop:=true;
    FileName:=sFileBmp;
    FrameCount:=nFrameCount;
    Left:=nX; Top:=nY;
    SndOnClicking:=sFileSndClick;
    SndHightLight:=sFileSndHiLight; // *.WAV sound file
    //  Set Hint
    HintOn:=sHintOn;
    HintOff:=sHintOff;
    // Set m_bEnabled, m_bChecked, m_bHLight
    m_bEnabled:=true;
    if  m_BtnType=bbChkBox then m_bChecked:=true
    else                        m_bChecked:=false;
    m_bHLight:=false;
    {$ifdef _DEBUG}
    if  m_BtnType=bbBtn then
        assert(FrameCount>=MAX_NUM_BTN_STATE);
    if  m_BtnType=bbChkBox then
        assert(FrameCount>=MAX_NUM_CHK_STATE);
    {$endif}
end; // ] TBtnAni.Init
//  Is a Check Button ?
function TBtnAni.IsChkBox: boolean;
begin // [
    Result:=(m_BtnType=bbChkBox);
end; // ] TBtnAni.IsChkBox
//  Is a Button ?
function TBtnAni.IsBtn: boolean;
begin // [
    Result:=(m_BtnType=bbBtn);
end; // ] TBtnAni.IsBtn
//  -----------------------------------------------------------------------------
//@ TBtnAni.MouseDown/TBtnAni.OnBtnUp/TBtnAni.OnBtnMove
//  -   procedure TBtnAni.MouseDown(Button: TMouseButton;
//                                    Shift: TShiftState;
//                                    X, Y: Integer);
//      For bbChkBox --> invoke OnClick callback event
//  -   procedure TBtnAni.MouseUp  (Button: TMouseButton;
//                                    Shift: TShiftState;
//                                    X, Y: Integer): boolean;
//      For bbBtn --> invoke OnClick callback event
//  -   procedure TBtnAni.MouseMove(Shift: TShiftState;
//                                    X, Y: Integer); override;
//  -----------------------------------------------------------------------------
procedure TBtnAni.MouseDown(Button: TMouseButton;
                              Shift: TShiftState;
                              X, Y: Integer);
begin // [
    if  Button<>mbLeft then exit;
    if  not IsEnabled then exit;
    if  not IsPtInControl(X, Y) then exit;
    //  for a Grouping Down or Checked Btn --> mouse message won't take effect
    if  (Down or Checked) and (_SameGroupIndexSiblingCount>0) then exit;
    if  Play=true then Play:=false;
    //  Callback TControl.OnMouseDown Event
    inherited MouseDown(Button, Shift, X, Y);
    m_bDragging:=true;
    s_bOnDragging:=true;
    // For Btn --> Do nothing
	if  m_BtnType=bbBtn then begin // [
        Frame:=bbBtnDn;
	end; // ] if m_BtnType=bbBtn
    // For Check Box --> Update Checked Status and invoke OnClick event
	if  m_BtnType=bbChkBox then begin // [
        if  GroupIndex=0 then
            Checked:=not Checked
        else
            UpdateExclusive;
        //  Callback TControl.OnClick Event
        _PlaySnd(m_sFileSndClicking); // Play Clicking Sound
        Click;
	end; // ] if m_BtnType=bbChkBox
end; // ] TBtnAni.MouseDown
procedure TBtnAni.MouseUp(Button: TMouseButton;
                           Shift: TShiftState;
                           X, Y: Integer);
begin // [
    if  Button<>mbLeft then exit;
    if  not IsEnabled then exit;
    m_bDragging:=false;
    s_bOnDragging:=false;
    //  for a Grouping Down or Checked Btn --> mouse message won't take effect
    if  (Down or Checked) and (_SameGroupIndexSiblingCount>0) then exit;
    //  Callback TControl.OnMouseUp Event
    inherited MouseUp(Button, Shift, X, Y);
    //  For Btn --> DrawBtnUp and invoke OnClick event
	if  m_BtnType=bbBtn then begin // [
        DrawBtnUp; // Draw BtnUp appearance
        if  IsPtInControl(X, Y) then begin
            UpdateExclusive;
            _PlaySnd(m_sFileSndClicking); // Play Clicking Sound
            Click; //  Call back TControl.OnClick Event
        end;
        (***
        if  GroupIndex=0 then begin // [ Non-GroupIndex Button
            DrawBtnUp; // Draw BtnUp appearance
            if  IsPtInControl(X, Y) then begin
                _PlaySnd(m_sFileSndClicking); // Play Clicking Sound
                Click; //  Call back TControl.OnClick Event
            end;
        end // ] Non-GroupIndex Button
        else begin // [ GroupIndex Buttons
            if  IsPtInControl(X, Y) then begin
                UpdateExclusive;
                _PlaySnd(m_sFileSndClicking); // Play Clicking Sound
                Click; //  Call back TControl.OnClick Event
            end;
        end; // ] GroupIndex Buttons
        ***)
	end; // ] if m_BtnType=bbBtn
    // For Check Box --> Do nothing
	if  m_BtnType=bbChkBox then begin // [
	end; // ] if m_BtnType=bbChkBox
end; // ] TBtnAni.MouseUp
procedure TBtnAni.MouseMove(Shift: TShiftState;
                              X, Y: Integer);
begin // [
    if  not IsEnabled then exit;
    if  not IsPtInControl(X, Y) then exit;
    //  for a Grouping Down or Checked Btn --> mouse message won't take effect
    if  (Down or Checked) and (_SameGroupIndexSiblingCount>0) then exit;
    inherited MouseMove(Shift, X, Y);
end; // ] TBtnAni.MouseMove
//  -----------------------------------------------------------------------------
//@ TBtnAni.DrawBtnUp;
//  procedure TBtnAni.DrawBtnUp;
//  Draw Btn Up appearance
//  -   bbBtn:      set Frame as bbBtnUp
//  -   bbChkBox:   set Frame as bbChkBoxOn/bbChkBoxOff
//  -----------------------------------------------------------------------------
procedure TBtnAni.DrawBtnUp;
begin // [
    if  Play=true then Play:=false;
    m_bHLight:=false;
	if  m_BtnType=bbBtn then begin // [
		if  m_bEnabled then begin
			Frame:=bbBtnUp;
		end
		else begin // [ Disable
			Frame:=bbBtnDisable;
		end; // ] Disable
		exit;
	end; // ] if m_BtnType=bbBtn
	if  m_BtnType=bbChkBox then begin // [
		if  m_bEnabled then begin
            if  IsChecked then Frame:=bbChkBoxOn // Check On
            else               Frame:=bbChkBoxOff;// Check Off
		end
		else begin // [ Disable
            if  IsChecked then Frame:=bbChkBoxOnDisable  // Check On
            else               Frame:=bbChkBoxOffDisable;// Check Off
		end; // ] Disable
		exit;
	end; // ] if m_BtnType=bbChkBox
end; // ] TBtnAni.DrawBtnUp
//  -----------------------------------------------------------------------------
//@ TBtnAni.IsEnabled/TBtnAni.SetEnabled
//  procedure TBtnAni.SetEnabled(Value: boolean);
//  function TBtnAni.IsEnabled : boolean;
//  Set/Get Btn Enabled/Disabled status
//  -----------------------------------------------------------------------------
procedure TBtnAni.SetEnabled(Value: boolean);
var
    bForceRefresh: boolean;
begin // [
    bForceRefresh:=false;
    if  Value<>m_bEnabled then bForceRefresh:=true;
    if  Play=true then Play:=false;
    //m_bHLight:=false;
    m_bEnabled:=Value;
	if  m_BtnType=bbBtn then begin // [
        if  IsHLight then exit;
		if  Value then begin
			Frame:=bbBtnUp;
		end
		else begin // [ Disable
			Frame:=bbBtnDisable;
		end; // ] Disable
        if(bForceRefresh) then Refresh;
		exit;
	end; // ] if m_BtnType=bbBtn
	if  m_BtnType=bbChkBox then begin // [
        if  IsHLight then exit;
		if  Value then begin
            if  IsChecked then Frame:=bbChkBoxOn // Check On
            else               Frame:=bbChkBoxOff;// Check Off
		end
		else begin // [ Disable
            if  IsChecked then Frame:=bbChkBoxOnDisable  // Check On
            else               Frame:=bbChkBoxOffDisable;// Check Off
		end; // ] Disable
        if(bForceRefresh) then Refresh;
		exit;
	end; // ] if m_BtnType=bbChkBox
    inherited Enabled:=Value;
end; // ] TBtnAni.SetEnabled
//  -----------------------------------------------------------------------------
//@ TBtnAni.IsEnabled
//  function TBtnAni.IsEnabled : boolean;
//  Get Btn Enabled/Disabled
//  -----------------------------------------------------------------------------
function TBtnAni.IsEnabled : boolean;
begin // [
    Result:=m_bEnabled;
end; // ] TBtnAni.IsEnabled
//  -----------------------------------------------------------------------------
//@ TBtnAni.SetChecked/TBtnAni.IsChecked
//  procedure TBtnAni.SetChecked(Value: boolean);
//  function TBtnAni.IsChecked : boolean;
//  Set/Get Btn Checked/UnChecked status
//  -----------------------------------------------------------------------------
function TBtnAni.IsChecked : boolean;
begin // [
    if  m_BtnType<>bbChkBox then begin
        Result:=false; exit;
    end;
    Result:=m_bChecked;
end; // ]  TBtnAni.IsChecked
procedure TBtnAni.SetChecked(Value: boolean);
var
    bForceRefresh: boolean;
begin // [
    bForceRefresh:=false;
    if  Value<>m_bChecked then bForceRefresh:=true;
    if  Play=true then Play:=false;
    m_bHLight:=false;
    if  m_BtnType<>bbChkBox then exit; // if !bbChkBox --> CANNOT SetChecked
    if  not IsEnabled then exit; // no Set Checked if !Enabled
    if  m_bChecked=Value then exit; // no need to update
    //  Toggle Checked
    if  Value then Frame:=bbChkBoxOn
    else           Frame:=bbChkBoxOff;
    m_bChecked:=Value;
    if( bForceRefresh) then Refresh;
    // if chk btn has two hint string --> toggle it
    if( (m_sHintOn<>'') and (m_sHintOff<>'')) then begin
        if m_bChecked then Hint:=m_sHintOn
        else               Hint:=m_sHintOff;
    end;
    UpdateExclusive;
end; // ] TBtnAni.SetChecked
//  -----------------------------------------------------------------------------
//@ TBtnAni.SetHLight/TBtnAni.IsHLight
//  procedure TBtnAni.SetHLight;
//  Set Button High Light
//  -   bbBtn
//      if  FrameCount>4 --> Play From 4~FrameCount-1
//      if  FrameCount=4 --> set Frame bbBtnHLight
//  -   bbChkBox
//      if  FrameCount>6 --> Play From 6~FrameCount-1
//      if  FrameCount=6 -->
//          if  Frame<=bbChkBoxOffDisable
//          --> Frame=bbChkBoxOffHLight;
//          if  Frame<=bbChkBoxOffDisable
//          --> Frame=bbChkBoxOffHLight;
//  -----------------------------------------------------------------------------
function TBtnAni.IsHLight: boolean;
begin // [
    Result:=m_bHLight;
end; // ] TBtnAni.IsHLight
procedure TBtnAni.SetHLight;
begin // [
    if  not IsEnabled then exit; // no Set High Light if !Enabled
    if  IsHLight then exit; // no duplicated set High Light
    m_bHLight:=true;
    if  m_BtnType=bbBtn then begin
        // if FrameCount>MAX_NUM_BTN_STATE --> Play From StartFrame~StopFrame
        // else                            --> set Frame bbBtnHLight
        if  FrameCount>MAX_NUM_BTN_STATE then begin
            Play:=true;
        end
        else begin
            Frame:=bbBtnHLight;
        end;
    end;
    if  m_BtnType=bbChkBox then begin
        // if FrameCount>MAX_NUM_CHK_STATE --> Play From StartFrame~StopFrame
        // else                            --> set Frame bbChkBoxOnHLight/bbChkBoxOffHLight
        if  FrameCount>MAX_NUM_CHK_STATE then begin
            Play:=true;
        end
        else begin
            if  IsChecked then Frame:=bbChkBoxOnHLight
            else               Frame:=bbChkBoxOffHLight;
        end;
    end;
    _PlaySnd(m_sFileSndHiLight); // Play High Light Sound
end; // ] TBtnAni.SetHLight
procedure TBtnAni._SetHintOn(sHintOn: string);
begin // [
    if  IsStrNonsense(sHintOn) then begin
        sHintOn:='';
        ShowHint:=False
    end
    else
        ShowHint:=True;
    Hint:=sHintOn;
    m_sHintOn:=sHintOn;
end; // ] TBtnAni._SetHintOn
procedure TBtnAni._SetHintOff(sHintOff: string);
begin // [
    if  IsStrNonsense(sHintOff) then
        sHintOff:='';
    m_sHintOff:=sHintOff;
end; // ] TBtnAni._SetHintOff
procedure TBtnAni._PlaySnd(sFile: string);
begin // [
    //  Play High Light Sound
    if  IsSndOn then begin // Play Voice Sound
        if  (sFile<>'') then begin
            sndPlaySound(PChar(sFile), SND_ASYNC);
        end;// ] if s_bSndOn and(sFile<>'')
    end; // ] case s_bSndOn of
end; // ] TBtnAni._PlaySnd
//  ------------------------------------------------------------------------
//@ TBtnAni.CMMouseEnter
//  procedure TBtnAni.CMMouseEnter(var Message: TMessage);
//  On Mouse Enter Control --> Enable High Light state
//  ------------------------------------------------------------------------
procedure TBtnAni.CMMouseEnter(var Message: TMessage);
begin // [
    inherited;
    if  not IsEnabled then exit;
    if  IsHLight then exit;
    //  for a Grouping Down or Checked Btn --> mouse message won't take effect
    if  (Down or Checked) and (_SameGroupIndexSiblingCount>0) then exit;
    //  if one of TBtnAni is dragging --> no set High Light of any others
    if  m_bDragging then exit;
    //if  s_bOnDragging then exit;
    SetHLight;
    if  Assigned(_OnMouseEnter) then _OnMouseEnter(Self);
end; // ] TBtnAni.CMMouseEnter
//  ------------------------------------------------------------------------
//@ TBtnAni.CMMouseLeave
//  procedure TBtnAni.CMMouseLeave(var Message: TMessage);
//  On Mouse Leave Control --> Disable High Light state
//  ------------------------------------------------------------------------
procedure TBtnAni.CMMouseLeave(var Message: TMessage);
begin // [
    inherited;
    if  not IsEnabled then exit;
    if  not IsHLight then exit;
    //  for a Grouping Down or Checked Btn --> mouse message won't take effect
    if  (Down or Checked) and (_SameGroupIndexSiblingCount>0) then exit;
    //  if current object is clicked and dragging
    //  when mouse leave object --> no Draw Btn Up of itself
    if  m_bDragging then exit;
    DrawBtnUp;
    if  Assigned(_OnMouseLeave) then _OnMouseLeave(Self);
end; // ] TBtnAni.CMMouseLeave
//  -----------------------------------------------------------------------------
//@ TBtnAni.Click;
//  procedure TBtnAni.Click;
//  TBtnAni override TControl.Click operation, this function is essential
//  if your component need take place of when happened Click operation
//  moveover the constructor in TImageAni.Create of controlStyle should be set as
//  ControlStyle:=[csCaptureMouse, csOpaque, csDoubleClicks]; which will no
//  longer set ControlState as csClicked
//  Note:   if (csClicked in ControlState) the unexpected another OnClick event
//          will be raising by TControl
//  -----------------------------------------------------------------------------
procedure TBtnAni.Click;
begin
    inherited Click;
end;
//  -----------------------------------------------------------------------------
//@ TBtnAni._SameGroupIndexSiblingCount
//  function TBtnAni._SameGroupIndexSiblingCount: integer;
//  980909
//  Get Button Same GroupIndex Sibling Count, not include itself
//  For a group Buttons, if one button is Down or Checked, then
//  the mouse event will be no longer effect
//  -----------------------------------------------------------------------------
function TBtnAni._SameGroupIndexSiblingCount: integer;
var
    Btn: TBtnAni; // Btn of Same GroupIndex
    i: integer;
    nSiblingCount: integer;
begin // [
    Result:=0;
    if  GroupIndex=0 then exit;
    nSiblingCount:=0;
    for i:=0 to Parent.ControlCount-1 do begin // [
        if  Parent.Controls[i] is TBtnAni then begin
            Btn:=TBtnAni(Parent.Controls[i]);
            if  Btn.GroupIndex<>GroupIndex then continue;
            if  Btn.BtnType<>BtnType then continue;
            if  Btn=Self then continue;
            inc(nSiblingCount);
        end;
    end; // ] for
    Result:=nSiblingCount;
end; // ] TBtnAni._SameGroupIndexSiblingCount
//  -----------------------------------------------------------------------------
//@ TBtnAni.UpdateExclusive;
//  procedure TBtnAni.UpdateExclusive;
//  980909
//  Update Same GroupIndex Sibling if GroupIndex<>0
//  -   For bbBtn
//      if one bbBtn Click --> Down current Btn and Up previous
//      Down Btn
//  -   For bbChkBox
//      if one bbChkBox Click --> SetChk of current Btn and Uncheck of
//      previous Checked Btn
//  -   Note that
//      1.  Same group index TBtnAni should have the same type
//      2.  Btn has no Same gorup sibling --> no Toggle Object itself
//  -----------------------------------------------------------------------------
{$J+} // Enable type constant
procedure TBtnAni.UpdateExclusive;
const
    s_bCanUpdateExclusive: boolean=TRUE;
var
    Btn: TBtnAni; // Btn of Same GroupIndex
    i: integer;
    nSiblingCount: integer;
begin // [
    if  not s_bCanUpdateExclusive then exit;
    if  GroupIndex=0 then exit;
    if  IsLoading(Self) then exit;
    nSiblingCount:=0;
    s_bCanUpdateExclusive:=FALSE; // Avoid Rescursive UpdateExclusive
    try // [
        for i:=0 to Parent.ControlCount-1 do begin // [
            if  Parent.Controls[i] is TBtnAni then begin
                Btn:=TBtnAni(Parent.Controls[i]);
                if  Btn.GroupIndex<>GroupIndex then continue;
                if  Btn.BtnType<>BtnType then continue;
                if  Btn=Self then continue;
                if  m_BtnType=bbBtn then Btn.Down:=FALSE;// Up previous Down Btn of Same Group
                if  m_BtnType=bbChkBox then Btn.Checked:=FALSE;// Uncheck previous Checked Btn of Same Group
                inc(nSiblingCount);
            end;
        end; // ] for
        //  if Single Group Btn --> Toggle Object itself
        if  nSiblingCount=0 then begin
            if  m_BtnType=bbBtn then Down:=not Down;
            if  m_BtnType=bbChkBox then Checked:=not Checked;
        end
        //  if Btn has Same gorup sibling --> Toggle Object itself
        else begin
            if  m_BtnType=bbBtn then Down:=TRUE; // Down current Clicked bbBtn
            if  m_BtnType=bbChkBox then Checked:=TRUE; // Checked current Clicked bbChkBox
        end; // if  nSiblingCount>0
    finally
        s_bCanUpdateExclusive:=TRUE;
    end; // ] try...finally
end; // ] UpdateExclusive
//  ------------------------------------------------------------------------
//@ TBtnAni.SetGroupIndex
//  proecedure TBtnAni.SetGroupIndex(Value: integer);
//  980909
//  -   not allow in run time
//  -   Set GroupIndex to determine how the button behaves when clicked.
//  -   When GroupIndex is 0, the button behaves independently of all other
//      buttons on the form. When the user clicks such a speed button,
//      the button appears pressed (in its clicked state) and then returns
//      to its normal up state when the user releases the mouse button.
//  -   When GroupIndex is greater than 0, the button remains selected
//      (in its down state) when clicked by the user. When the user clicks
//      a selected button, it returns to the up state, unless AllowAllUp is
//      False.
//  -   Setting the GroupIndex property of a single speed button to a value
//      greater than 0 causes the button to behave as a two-state button when
//      AllowAllUp is True.
//  -   Speed buttons with the same GroupIndex property value (other than 0),
//      work together as a group. When the user clicks one of these buttons,
//      it remains selected until the user clicks another speed button belonging
//      to the same group. Speed buttons used in this way can present mutually
//      exclusive choices to the user.
//      Note: AllowAllUp not implement yet
//  ------------------------------------------------------------------------
procedure TBtnAni.SetGroupIndex(Value: integer);
begin // [
    //  not allow in run time
    if  FGroupIndex=Value then exit;
    if  FGroupIndex=0 then DrawBtnUp;
    FGroupIndex:=Value;
    if  not (csDesigning in ComponentState) then exit;
    UpdateExclusive;
end; // ] TBtnAni.SetGroupIndex
//  ------------------------------------------------------------------------
//@ TBtnAni.SetDown
//  proecedure TBtnAni.SetDown(Value: boolean);
//  980909
//  Btn can Down only if, Btn is bbBtn and GroupIndex<>0
//  -   When GroupIndex is greater than 0, set Down to True to select a button.
//      When the user clicks on a button in the unselected (up) state, the
//      button is selected and Down is set to True.
//  -   When the user clicks on a button in the selected (down) state,
//      if AllowAllUp is True, the button becomes unselected and Down is set
//      to False.
//      Note: AllowAllUp not implement yet
//  ------------------------------------------------------------------------
procedure TBtnAni.SetDown(Value: boolean);
begin // [
    if  FDown=Value then exit;
    if  BtnType<>bbBtn then exit;
    if  GroupIndex=0 then exit;
    FDown:=Value;
    if  FDown then
        Frame:=bbBtnDn
    else
        DrawBtnUp;
    UpdateExclusive;
end; // ] TBtnAni.SetDown
initialization
    s_bOnDragging:=false;
end.
