//  ------------------------------------------------------------------------
//@ ImageAni.pas
//  TImageAni VCL Component
//  Copyright (c) Cole Joe Taiwan 1998 All rights reserved.
//  email: kchou@adc.iii.org.tw
//  -   Description
//      TImageAni is a component combine the properties of TAnimate and
//      TImage, which load Glyph from BMP file other than TAnimate from AVI
//      user can specify FrameCount, StartFrame, StopFrame and Loop then
//      issue Play:=TRUE command for looping play. Stretch and Transparent
//      supported.
//  -   980910
//      First implement this class, modified by referencing TImage
//  -   object description
//      Animation TImage, which support looping/reverse play
//      eg. ImageAni1.FrameCount:=10;
//          ImageAni1.Transparent:=TRUE;
//          ImageAni1.LoadFromFile('foo.bmp');
//          ImageAni1.Play:=TRUE;
//  -   properties decription
//      Interval:       frequently of playing frame
//      FrameCount:     number of frames that can be displayed by the
//                      animation control.
//      StartFrame:     Start Frame
//      StopFrame:      Stop Frame
//      Frame:          Current Frame
//      FileName:       Image File Name
//      Play:           Play Animation or not
//      Reverse:        Reverse Play or not
//      Loop:           Looping Play or not
//      OnChangeFrame:  Notify Event when changing frame
//      Stretch:        Strectch showing
//      Transparent:    Transparent showing. Default Tranparent color is
//                      located on left-bottom of Image
//  ------------------------------------------------------------------------
unit ImageAni;
interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;
type
  TImageAni=class(TGraphicControl)
  private
    m_nStartFrame: integer; // Playing Frame Begin
    m_nStopFrame: Integer; // Playing Frame End
    m_nFrameCount: integer;
    m_nCurFrame: Integer;
    m_Timer: TTimer;
    m_sFileName: string;
    m_nInterval: integer; // Playing interval (msec)
    m_bLooping: boolean;     // looping playing
    m_bReversePlay: boolean; // reverse playing
    m_bIsPlaying: boolean;
    _OnChangeFrame: TNotifyEvent;
    FPicture: TPicture;
    FOnProgress: TProgressEvent;
    FAutoSize: Boolean;
    FStretch: Boolean;
    FCenter: Boolean;
    FIncrementalDisplay: Boolean;
    FTransparent: Boolean;
    FDrawing: Boolean;
    procedure   _SetTransparentColor(Value: TColor);
    function    _GetTransparentColor: TColor;
    function    GetCanvas: TCanvas;
    procedure   PictureChanged(Sender: TObject);
    procedure   SetAutoSize(Value: Boolean);
    procedure   SetCenter(Value: Boolean);
    procedure   SetPicture(Value: TPicture);
    procedure   SetStretch(Value: Boolean);
    procedure   SetTransparent(Value: Boolean);
    procedure   _ChkStop;
    procedure   _NewTimer(nInterval: integer);
    procedure   _DelTimer;
  protected
    procedure   _SetFrameBeg(Value: Integer);
    procedure   _SetFrameEnd(Value: Integer);
    procedure   _SetCurFrame(Value: Integer);
    procedure   _SetTotalFrame(Value: Integer); virtual;
    procedure   _SetInterval(Value: integer);
    procedure   _SetPlay(Value: boolean);
    procedure   _TimeHit(Sender: TObject);
    function    DestRect: TRect;
    function    DoPaletteChange: Boolean;
    function    GetPalette: HPALETTE; override;
    procedure   Paint; override;
    procedure   Progress(Sender: TObject;
                         Stage: TProgressStage;
                         PercentDone: Byte;
                         RedrawNow: Boolean;
                         const R: TRect;
                         const Msg: string); dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    property    Canvas: TCanvas read GetCanvas;
    procedure   LoadFromFile(const Value: string);
    function    IsPtInControl(x, y: integer): boolean;
  published
    property Interval: integer              read m_nInterval
                                            write _SetInterval;
    property FrameCount: integer            read m_nFrameCount
                                            write _SetTotalFrame
                                            default 1;
    property StartFrame: integer            read m_nStartFrame
                                            write _SetFrameBeg
                                            default 0;
    property StopFrame: integer             read m_nStopFrame
                                            write _SetFrameEnd
                                            default 0;
    property Frame: Integer                 read m_nCurFrame
                                            write _SetCurFrame;
    property FileName: string               read m_sFileName
                                            write LoadFromFile;
    property Play: boolean                  read m_bIsPlaying
                                            write _SetPlay;
    property Reverse: boolean               read m_bReversePlay
                                            write m_bReversePlay;
    property Loop: boolean                  read m_bLooping
                                            write m_bLooping
                                            default True;
    property OnChangeFrame:TNotifyEvent     read _OnChangeFrame
			                                write _OnChangeFrame;
    //////////////////////////////
    property Align;
    property AutoSize: Boolean              read FAutoSize
                                            write SetAutoSize
                                            default False;
    property Center: Boolean                read FCenter
                                            write SetCenter
                                            default False;
    property DragCursor;
    property DragMode;
    property Enabled;
    property IncrementalDisplay: Boolean    read FIncrementalDisplay
                                            write FIncrementalDisplay
                                            default False;
    property ParentShowHint;
    property Picture: TPicture              read FPicture
                                            write SetPicture;
    property PopupMenu;
    property ShowHint;
    property Stretch: Boolean               read FStretch
                                            write SetStretch
                                            default False;
    property Transparent: Boolean           read FTransparent
                                            write SetTransparent
                                            default False;
    property TransparentColor: TColor       read _GetTransparentColor
                                            write _SetTransparentColor;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnProgress: TProgressEvent     read FOnProgress
                                            write FOnProgress;
    property OnStartDrag;
  end;
{ TImageAni }
implementation
{$R ImageAni.res}
constructor TImageAni.Create(AOwner: TComponent);
begin
  	inherited Create(AOwner);
  	ControlStyle:=ControlStyle + [csReplicatable];
  	FPicture:=TPicture.Create;
  	FPicture.OnChange:=PictureChanged;
  	FPicture.OnProgress:=Progress;
    /////////////
    m_sFileName:='';
    m_Timer:=nil;
    Width:=30;
    Height:=30;
    m_nFrameCount:= 1;
    m_nInterval:=200;
    //  980908 Note: DON'T use
    //  ControlStyle:=ControlStyle +[csCaptureMouse, csOpaque, csDoubleClicks]; // 980908
    //  because ControlState CANNOT be csClicked
    ControlStyle:=[csReplicatable, csCaptureMouse, csOpaque, csDoubleClicks];
    //ControlStyle:=[csCaptureMouse, csDoubleClicks];
    m_bLooping:=True;
    //Transparent:=TRUE;
    m_sFileName:='';
    m_nStartFrame:=0; // Playing Frame Begin
    m_nStopFrame:=0; // Playing Frame End
end;
destructor TImageAni.Destroy;
begin
    _DelTimer;
  	FPicture.Free;
  	inherited Destroy;
end;
function TImageAni.GetPalette: HPALETTE;
begin
  	Result:=0;
  	if 	FPicture.Graphic<>nil then
    	Result:=FPicture.Graphic.Palette;
end;
function TImageAni.DestRect: TRect;
var
    nWidth, nHeight: integer; // Frame bitmap size
begin // [
    nWidth:=Picture.Width div FrameCount;
    nHeight:=Picture.Height;
  	if 	Stretch then
    	Result:=ClientRect
  	else if Center then
    	Result:=Bounds((Width - nWidth) div 2,
                       (Height - nHeight) div 2,
      				   nWidth,
                       nHeight)
  	else
    	Result:=Rect(0, 0, nWidth, nHeight);
end; // ] DestRect
function TImageAni.DoPaletteChange: Boolean;
var
  	ParentForm: TCustomForm;
  	Tmp: TGraphic;
begin // [
  	Result:=False;
  	Tmp:=Picture.Graphic;
  	if 	Visible and
        (not (csLoading in ComponentState)) and
        (Tmp<>nil) and
    	(Tmp.PaletteModified) then begin
    	if 	(Tmp.Palette=0) then
      		Tmp.PaletteModified:=False
    	else begin // [
      		ParentForm:=GetParentForm(Self);
      		if 	Assigned(ParentForm) and
                ParentForm.Active and
                Parentform.HandleAllocated then begin
        		if 	FDrawing then
          			ParentForm.Perform(wm_QueryNewPalette, 0, 0)
        		else
          			PostMessage(ParentForm.Handle, wm_QueryNewPalette, 0, 0);
        		Result:=True;
        		Tmp.PaletteModified:=False;
      		end;
    	end; // ] if(Tmp.Palette<>0)
  	end;
end; // ] TImageAni.DoPaletteChange
procedure TImageAni.Progress(Sender: TObject;
                             Stage: TProgressStage;
  	                         PercentDone: Byte;
                             RedrawNow: Boolean;
                             const R: TRect;
                             const Msg: string);
begin // [
  	if  FIncrementalDisplay and RedrawNow then begin
    	if DoPaletteChange then Update
    	else Paint;
  	end;
  	if Assigned(FOnProgress) then FOnProgress(Sender, Stage, PercentDone, RedrawNow, R, Msg);
end; // ] TImageAni.Progress
function TImageAni.GetCanvas: TCanvas;
var
  	Bitmap: TBitmap;
begin // [
  	if  Picture.Graphic=nil then begin
    	Bitmap:=TBitmap.Create;
    	try
      		Bitmap.Width:=Width;
      		Bitmap.Height:=Height;
      		Picture.Graphic:=Bitmap;
    	finally
      		Bitmap.Free;
    	end;
  	end;
  	if 	Picture.Graphic is TBitmap then
    	Result:=TBitmap(Picture.Graphic).Canvas
  	else
    	raise EInvalidOperation.Create('Image Canvas need Bitmap');
end; // ] TImageAni.GetCanvas
procedure TImageAni.SetAutoSize(Value: Boolean);
begin
  	FAutoSize:=Value;
    if  FAutoSize then
        FStretch:=FALSE;
    PictureChanged(Self);
end;
procedure TImageAni.SetCenter(Value: Boolean);
begin
  	if 	Value=FCenter then exit;
   	FCenter:=Value;
   	PictureChanged(Self);
end;
procedure TImageAni.SetPicture(Value: TPicture);
begin // [
  	FPicture.Assign(Value);
    ///////////
    Height:=Picture.Bitmap.Height;
    if  Height=0 then Height:=30;
    Width:=Picture.Width div FrameCount;
    if  Width=0 then Width:=30;
    //Invalidate;
end; // ] TImageAni.SetPicture
procedure TImageAni.SetStretch(Value: Boolean);
begin
  	if 	Value=FStretch then exit;
   	FStretch:=Value;
   	PictureChanged(Self);
end;
procedure TImageAni.SetTransparent(Value: Boolean);
begin
  	if 	Value=FTransparent then exit;
   	FTransparent:=Value;
   	PictureChanged(Self);
end;
procedure TImageAni.PictureChanged(Sender: TObject);
var
  	G: TGraphic;
begin
  	if 	AutoSize and (Picture.Width>0) and (Picture.Height>0) then
    	SetBounds(Left, Top, Picture.Width div FrameCount, Picture.Height);
  	G:=Picture.Graphic;
  	if 	G<>nil then begin
    	if 	not ((G is TMetaFile) or (G is TIcon)) then
      		G.Transparent:= FTransparent;
    	if 	(not G.Transparent) and
    		(Stretch or (G.Width >= Width) and (G.Height >= Height)) then
      		ControlStyle:=ControlStyle + [csOpaque]
    	else
      		ControlStyle:=ControlStyle - [csOpaque];
    	if 	DoPaletteChange and FDrawing then Update;
  	end
  	else
  		ControlStyle:=ControlStyle - [csOpaque];
  	if 	not FDrawing then Invalidate;
end;
/////////////////////////////////////////////////
procedure TImageAni._NewTimer(nInterval: integer);
begin
    _DelTimer;
    m_Timer:=TTimer.Create(Self);
    m_Timer.Interval:=nInterval;
    m_Timer.OnTimer:=_TimeHit;
end;
procedure TImageAni._DelTimer;
begin //[
    if  m_Timer=nil then exit;
    m_Timer.Free; m_Timer:=nil;
end; // ] TImageAni._DelTimer
procedure TImageAni.LoadFromFile(const Value: string);
begin // [
    if  not FileExists(Value) then begin
        Picture:=nil; m_sFileName:=''; exit;
    end;
    if  CompareText(Value, m_sFileName)=0 then exit; // EqualNoCase
    m_sFileName:=Value;
    try // [
        Picture.LoadFromFile(Value);
        Height:=Picture.Bitmap.Height;
        if  Height=0 then Height:=30;
        Width:=Picture.Width div FrameCount;
        if  Width=0 then Width:=30;
   	    PictureChanged(Self);
    except
        m_sFileName:='';
        raise;
    end; // ] try...except
    //Paint;
end; // ] TImageAni.LoadFromFile
procedure TImageAni._SetInterval(Value: Integer);
begin // [
    if  Value=m_nInterval then exit;
    if  Value<=0 then Value:=0;
    if  m_bIsPlaying and (Value>0) then begin
        _NewTimer(Value);
    end;
    m_nInterval:=Value;
end; // ] TImageAni._SetInterval
procedure TImageAni._SetPlay(Value: boolean);
begin // [
    if  Value=m_bIsPlaying then exit;
    m_bIsPlaying:=Value;
    if  not Value then 
        _DelTimer
    else if m_nInterval>0 then begin
        Frame:=m_nStartFrame; // Begin Playing from start
        _NewTimer(m_nInterval);
    end;
end; // ] TImageAni._SetPlay
procedure TImageAni._SetCurFrame(Value: Integer);
var
    nTmp: Integer;
begin // [
    if  Value<0 then
	    nTmp:=0
    else
	    nTmp:=Value Mod FrameCount;
    if  nTmp=m_nCurFrame then exit;
    m_nCurFrame:=nTmp;
    if Assigned(_OnChangeFrame) then _OnChangeFrame(Self);
    //PictureChanged(Self);
    Invalidate;
end; // ] TImageAni._SetCurFrame
// Added By Cole
procedure TImageAni._SetTotalFrame(Value: Integer);
begin // [
    if( (Value<=0) or (Value>1000)) then begin
	    MessageDlg('FrameCount out of range(1~999)', mtError, [mbOK], 0);
	    exit;
    end;
    m_nFrameCount:=Value;
    StartFrame:=0; StopFrame:=Value-1;
    if Picture.Bitmap.Empty then exit;
    Height:=Picture.Bitmap.Height;
    Width:=Picture.Width div FrameCount;
    m_nCurFrame:=-1;
    _SetCurFrame(0);
end; // ] TImageAni._SetTotalFrame
// Set Play Begin Frame
procedure TImageAni._SetFrameBeg(Value: Integer);
begin // [
    if Value<0 then Value:=0;
    if Value>StopFrame then Value:=StopFrame;
    m_nStartFrame:=Value;
end; // ] TImageAni._SetFrameBeg
// Set Play End Frame
procedure TImageAni._SetFrameEnd(Value: Integer);
begin // [
    if  Value>=FrameCount then Value:=FrameCount-1;
    if  Value<StartFrame then Value:=StartFrame;
    m_nStopFrame:=Value;
end; // ] TImageAni._SetFrameEnd
procedure TImageAni._ChkStop;
begin // [
    if  not m_bLooping then begin
        m_bIsPlaying:=False;
        _DelTimer;
    end;
end; // ] _ChkStop
procedure TImageAni._TimeHit(Sender: TObject);
var
    nToFrame: integer;
begin // [
    if  m_bReversePlay then begin
        nToFrame:=Frame-1;
        if nToFrame<StartFrame then nToFrame:=StopFrame;
        Frame:=nToFrame;
	    if m_nCurFrame=StartFrame then _ChkStop;
    end
    else begin
        nToFrame:=Frame+1;
        if  nToFrame>StopFrame then nToFrame:=StartFrame;
        Frame:=nToFrame;
	    if  m_nCurFrame=StopFrame then _ChkStop;
    end;
end; // ] TImageAni._TimeHit
//  Set m_TransparentColor by ref FTransparent and FPicture
function  TImageAni._GetTransparentColor: TColor;
begin // [
    Result:=clNone;
    if  not Transparent then exit;
    if  Picture.Bitmap.Empty then exit;
    Result:=Picture.Bitmap.TransparentColor;
end; // ] TImageAni._GetTransparentColor
procedure TImageAni._SetTransparentColor(Value: TColor);
begin // [
    Picture.Bitmap.TransparentColor:=Value;
end; // ] TImageAni._SetTransparentColor
//  -----------------------------------------------------------------------------
//@ TImageAni.IsPtInControl
//  function TImageAni.IsPtInControl(x, y: integer): boolean;
//  Check if Mouse Point is in Control
//  Note: the incomming mouse point (x,y) is w.r.t. TImageAni itself
//  -----------------------------------------------------------------------------
function TImageAni.IsPtInControl(x, y: integer): boolean;
begin // [
    Result:=false;
    if  x<0 then exit;
    if  y<0 then exit;
    if  x>=Width then exit;
    if  y>=Height then exit;
    if  Transparent then begin
        if  Canvas.Pixels[X,Y]=TransparentColor then exit;
    end;
    Result:=TRUE;
end; // ] TImageAni.IsPtInControl
////////////////////////////////////////////////
//  ------------------------------------------------------------------------
//@ TImageAni.Paint;
//  procedure TImageAni.Paint;
//  Copy TImageAni i-th Frame partial bitmap then draw on its
//  TGraphicControl.Canvas (represent as 'inherited.Canvas)
//  cf. TImageAni.Canvas is Picture.Bitmap.Canvas
//      inherited Canvas is TGraphicControl.Canvas
//  ------------------------------------------------------------------------
procedure _GetPartialPicture(PicDst: TPicture; rDst: TRect;
                             PicSrc: TPicture; rSrc: TRect);
begin // [
    PicDst.Bitmap.Width:=rDst.Right-rDst.Left;
    PicDst.Bitmap.Height:=rDst.Bottom-rDst.Top;
    //  Copy rSrc of PicSrc.Graphic to PicDst
    PicDst.Bitmap.Canvas.CopyRect(rDst, PicSrc.Bitmap.Canvas, rSrc);
    PicDst.Graphic.Transparent:=PicSrc.Graphic.Transparent;
end; // ] _GetPartialPicture
procedure TImageAni.Paint;
var
  	bSave: Boolean;
    rDst, rSrc : TRect;
    X : Integer;
    nWidthSrc: integer;
    nHeightSrc: integer;
    P: TPicture;
begin // [
  	if 	csDesigning in ComponentState then
    	with inherited Canvas do begin
      		Pen.Style:=psDash;
      		Brush.Style:=bsClear;
      		Rectangle(0, 0, Width, Height);
    	end;
    if  Picture.Bitmap.Empty then exit;
    if  FDrawing then exit;
  	bSave:=FDrawing;
  	FDrawing:=True;
    //  TPicture, TGraphic
  	try // [
        nWidthSrc:=Picture.Width div FrameCount;
        nHeightSrc:=Picture.Height;
        X:=nWidthSrc*Frame;
        rDst:=Rect(0,0, nWidthSrc, nHeightSrc);
        rSrc:=Rect(X,0, X+nWidthSrc, nHeightSrc);
        if  Stretch then begin // [
            //  ********************************************************
            //  Stretch && (Tranparent or !Tranparent) (Slowest way)
            //  Canvas.StretchDraw
            //  ********************************************************
            P:=TPicture.Create;
            try // [
                _GetPartialPicture(P, rDst, Picture, rSrc);
                inherited Canvas.StretchDraw(DestRect, P.Graphic);
            finally
                P.Free;
            end; // ] try...finally
        end // ] Stretch
        else begin // [ !Stretch
            //  ********************************************************
            //  !Stretch && !Transparent: (Fastest way)
            //  Canvas.CopyRect
            //  ********************************************************
            if  not Transparent then begin
                inherited Canvas.CopyRect(rDst, Picture.Bitmap.Canvas, rSrc);
            end
            //  ********************************************************
            //  !Stretch && Transparent:
            //  Canvas.Draw
            //  ********************************************************
            else begin // [
                P:=TPicture.Create;
                try // [
                    _GetPartialPicture(P, rDst, Picture, rSrc);
                    inherited Canvas.Draw(0,0, P.Graphic);
                finally
                    P.Free;
                end; // ] try...finally
            end; // ] !Stretch && Transparent
        end; // ] !Stretch
  	finally
    	FDrawing:=bSave;
  	end; // ] try...finally
end; // ] TImageAni.Paint
end.

